from __future__ import annotations

import typing

import polars as pl


def prefix_hex_series_to_binary(series: pl.Series) -> pl.Series:
    return series.str.slice(2).map_elements(
        bytes.fromhex,
        return_dtype=pl.datatypes.Binary,
    )


def raw_hex_series_to_binary(series: pl.Series) -> pl.Series:
    return series.map_elements(bytes.fromhex, return_dtype=pl.datatypes.Binary)


def binary_series_to_prefix_hex(series: pl.Series) -> pl.Series:
    return '0x' + binary_series_to_raw_hex(series)


def binary_series_to_raw_hex(series: pl.Series) -> pl.Series:
    as_hex = [(x.hex() if x is not None else None) for x in series.to_list()]
    return pl.Series(series.name, as_hex, pl.datatypes.Utf8)


def binary_columns_to_prefix_hex(
    df: pl.DataFrame, columns: typing.Sequence[str] | None = None
) -> pl.DataFrame:
    if columns is None:
        columns = df.select(pl.col(pl.datatypes.Binary)).columns
    return df.with_columns(
        [binary_series_to_prefix_hex(df[column]) for column in columns]
    )


def binary_columns_to_raw_hex(
    df: pl.DataFrame, columns: typing.Sequence[str] | None = None
) -> pl.DataFrame:
    if columns is None:
        columns = df.select(pl.col(pl.datatypes.Binary)).columns
    return df.with_columns(
        [binary_series_to_raw_hex(df[column]) for column in columns]
    )


def prefix_hex_columns_to_binary(
    df: pl.DataFrame, columns: typing.Sequence[str] | None = None
) -> pl.DataFrame:
    if columns is None:
        columns = []
        for column, dtype in zip(df.columns, df.dtypes):
            if dtype == pl.datatypes.Utf8:
                try:
                    bytes.fromhex(df[column][0][2:])
                    columns.append(column)
                except Exception:
                    continue
    return df.with_columns(
        [prefix_hex_series_to_binary(df[column]) for column in columns]
    )


def raw_hex_columns_to_binary(
    df: pl.DataFrame, columns: typing.Sequence[str] | None = None
) -> pl.DataFrame:
    if columns is None:
        columns = []
        for column, dtype in zip(df.columns, df.dtypes):
            if dtype == pl.datatypes.Utf8:
                try:
                    bytes.fromhex(df[column][0])
                    columns.append(column)
                except Exception:
                    continue
    return df.with_columns(
        [raw_hex_series_to_binary(df[column]) for column in columns]
    )



from __future__ import annotations

import typing

import polars as pl

from .. import _helpers


@pl.api.register_dataframe_namespace('evm')
class DataFrameEvm:
    def __init__(self, df: pl.DataFrame):
        self._df = df

    def binary_to_hex(
        self, columns: typing.Sequence[str] | None = None, prefix: bool = True
    ) -> pl.DataFrame:
        if prefix:
            return _helpers.binary_columns_to_prefix_hex(
                self._df, columns=columns
            )
        else:
            return _helpers.binary_columns_to_raw_hex(self._df, columns=columns)

    def hex_to_binary(
        self, columns: typing.Sequence[str] | None = None, prefix: bool = True
    ) -> pl.DataFrame:
        if prefix:
            return _helpers.prefix_hex_columns_to_binary(
                self._df, columns=columns
            )
        else:
            return _helpers.raw_hex_columns_to_binary(self._df, columns=columns)

    def binary_to_float(
        self,
        column_types: dict[str, str],
        replace: bool = False,
    ) -> pl.DataFrame:
        return _helpers.binary_df_to_float(
            df=self._df, column_types=column_types, replace=replace
        )



from __future__ import annotations

import typing

import polars as pl

from .. import _helpers


@pl.api.register_series_namespace('evm')
class SeriesEvm:
    def __init__(self, series: pl.Series):
        self._series = series

    def binary_to_hex(self, prefix: bool = True) -> pl.Series:
        if prefix:
            return _helpers.binary_series_to_prefix_hex(self._series)
        else:
            return _helpers.binary_series_to_raw_hex(self._series)

    def hex_to_binary(self, prefix: bool = True) -> pl.Series:
        if prefix:
            return _helpers.prefix_hex_series_to_binary(self._series)
        else:
            return _helpers.raw_hex_series_to_binary(self._series)

    def binary_to_float(self, raw_type: str) -> pl.Seties:
        return _helpers.binary_series_to_float(
            hex_series=self._series, raw_type=raw_type
        )

    def keccak(
        self,
        output: typing.Literal[
            'hex', 'binary', 'prefix_hex', 'raw_hex'
        ] = 'hex',
        text: bool = False,
    ) -> pl.Series:
        return self._series.map_elements(
            lambda datum: _helpers.keccak(datum, output=output, text=text)
        )



from __future__ import annotations

import typing

import polars as pl

from .. import _helpers


@pl.api.register_lazyframe_namespace('evm')
class LazyFrameEvm:
    def __init__(self, lf: pl.LazyFrame):
        self._lf = lf

    def binary_to_hex(
        self, columns: typing.Sequence[str] | None = None, prefix: bool = True
    ) -> pl.LazyFrame:
        if columns is None:
            columns = []
            for column, dtype in zip(self._lf.columns, self._lf.dtypes):
                if dtype == pl.datatypes.Binary:
                    columns.append(column)

        if prefix:
            return self._lf.with_columns(
                [
                    pl.lit('0x') + pl.col(column).bin.encode('hex')
                    for column in columns
                ]
            )
        else:
            return self._lf.with_columns(
                [pl.col(column).bin.encode('hex') for column in columns]
            )

    def hex_to_binary(
        self, columns: typing.Sequence[str] | None = None, prefix: bool = True
    ) -> pl.LazyFrame:
        if columns is None:
            columns = []
            for column, dtype in zip(self._lf.columns, self._lf.dtypes):
                if dtype == pl.datatypes.Binary:
                    columns.append(column)

        if prefix:
            return self._lf.with_columns(
                [
                    pl.col(column).slice(2).str.decode('hex')
                    for column in columns
                ]
            )
        else:
            return self._lf.with_columns(
                [pl.col(column).str.decode('hex') for column in columns]
            )

    def binary_to_float(
        self,
        column_types: dict[str, str],
        replace: bool = False,
    ) -> pl.LazyFrame:
        return _helpers.binary_df_to_float(
            df=self._lf, column_types=column_types, replace=replace
        )



from __future__ import annotations

import typing

import polars as pl

from .. import _helpers


@pl.api.register_expr_namespace('evm')
class ExprEvm:
    def __init__(self, expr: pl.Expr):
        self._expr = expr

    def binary_to_hex(self, prefix: bool = True) -> pl.Expr:
        if prefix:
            name = self._expr.meta.output_name()
            expr = pl.lit('0x') + self._expr.bin.encode('hex')
            return expr.alias(name)
        else:
            return self._expr.bin.encode('hex')

    def hex_to_binary(
        self, columns: typing.Sequence[str] | None = None, prefix: bool = True
    ) -> pl.Expr:
        if prefix:
            return self._expr.slice(2).str.decode('hex')
        else:
            return self._expr.str.decode('hex')

    def hex_to_float(self, raw_type: str) -> pl.Expr:
        return _helpers.hex_expr_to_float(self._expr, raw_type=raw_type)

    def keccak(
        self,
        output: typing.Literal[
            'hex', 'binary', 'prefix_hex', 'raw_hex'
        ] = 'hex',
        text: bool = False,
    ) -> pl.Expr:
        return self._expr.map_elements(
            lambda datum: _helpers.keccak(datum, output=output, text=text)
        )



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl

    T = typing.TypeVar('T', pl.DataFrame, pl.LazyFrame)


def binary_df_to_float(
    df: T,
    column_types: dict[str, str],
    replace: bool = False,
) -> T:
    # ) -> pl.DataFrame:
    """column_types in format {'col1': 'u256', 'col2': 'i128', ...}"""
    import polars as pl

    hex_columns = {}
    float_columns = {}
    for column, raw_type in column_types.items():
        # decide hex column
        column_dtype = df.collect_schema().get(column)
        if column_dtype == pl.Binary:
            hex_name = column + '_hex_tmp'
            hex_columns[hex_name] = pl.col(column).bin.encode('hex')
            hex_expr = pl.col(hex_name)
        elif column_dtype == pl.String:
            hex_name = column
            hex_columns = {}
            hex_expr = pl.col(hex_name)
        else:
            raise Exception('invalid column dtype:' + str(column_dtype))

        # rename
        if replace:
            float_name = column
        else:
            float_name = column + '_f64'

        # create float column
        float_columns[float_name] = hex_expr_to_float(
            hex_expr, raw_type=raw_type
        )

    return (
        df.with_columns(**hex_columns)
        .with_columns(**float_columns)
        .drop(*hex_columns.keys())
    )


def binary_series_to_float(hex_series: pl.Series, raw_type: str) -> pl.Series:
    import polars as pl

    name = hex_series.name
    if name is None:
        name = 'series'

    df = pl.DataFrame({name: hex_series})
    df = binary_df_to_float(df, {name: raw_type}, replace=True)
    return df[name]


def hex_expr_to_float(hex_expr: pl.Expr, raw_type: str) -> pl.Expr:
    import polars as pl

    # parse raw type
    raw_type = raw_type.lower()
    if raw_type.startswith('uint'):
        signed = False
        n_bits = int(raw_type[4:])
    elif raw_type.startswith('u'):
        signed = False
        n_bits = int(raw_type[1:])
    elif raw_type.startswith('int'):
        signed = True
        n_bits = int(raw_type[3:])
    elif raw_type.startswith('i'):
        signed = True
        n_bits = int(raw_type[1:])
    else:
        raise Exception()

    # build expression based on whether type is signed
    if signed:
        is_negative = hex_expr.str.slice(0, 2) > '7f'
        negative = _raw_hex_to_float(hex_expr, n_bits=n_bits, invert=True)
        positive = _raw_hex_to_float(hex_expr, n_bits=n_bits, invert=False)
        return pl.when(is_negative).then(negative).otherwise(positive)
    else:
        return _raw_hex_to_float(hex_expr, n_bits=n_bits, invert=False)


def _raw_hex_to_float(
    hex_expr: pl.Expr, *, n_bits: int, invert: bool
) -> pl.Expr:
    if n_bits % 8 != 0:
        raise Exception('n_bits must be divisible by 8')
    n_remaining = n_bits
    exprs = []
    while n_remaining > 0:
        chunk_start = int((n_bits - n_remaining) / 8)
        chunk_size = min(8, n_remaining)
        expr = _float_chunk(
            chunk_start,
            chunk_size,
            n_bits,
            invert=invert,
            hex_expr=hex_expr,
        )
        exprs.append(expr)
        n_remaining -= 8 * chunk_size
    expr = sum(exprs)  # type: ignore
    if invert:
        expr = -expr
    return expr


def _float_chunk(
    start_byte: int,
    n_chunk_bytes: int,
    total_bits: int,
    *,
    hex_expr: pl.Expr,
    invert: bool = False,
) -> pl.Expr:
    import polars as pl

    if n_chunk_bytes > 8:
        raise Exception('n_chunk_bytes must be <= 8')

    expr = (
        hex_expr.str.slice(2 * start_byte, 2 * n_chunk_bytes)
        .str.decode('hex')
        .bin.reinterpret(dtype=pl.UInt64, endianness='big')
    )

    if invert:
        max_value = 2 ** int(8 * n_chunk_bytes) - 1
        expr = pl.lit(max_value, dtype=pl.UInt64) - expr

    factor = total_bits - 8.0 * (start_byte + n_chunk_bytes)
    expr = expr.cast(pl.Float64) * (2**factor)

    return expr



from __future__ import annotations

import typing


def keccak(
    data: str | bytes,
    output: typing.Literal['hex', 'binary', 'prefix_hex', 'raw_hex'] = 'hex',
    text: bool = False,
) -> str | bytes:
    from Crypto.Hash import keccak as f_keccak

    # encode text as bytes
    if text:
        if isinstance(data, str):
            data = data.encode()
        else:
            raise Exception('not str, cannot use text=True')

    # decode hex into bytes
    if isinstance(data, str):
        try:
            if data.startswith('0x'):
                data = data[2:]
            data = bytes.fromhex(data)
        except ValueError:
            raise Exception('for text data, use text=True')

    # perform hash
    as_binary = f_keccak.new(digest_bits=256, data=data).digest()

    # convert to output format
    if output == 'binary':
        return as_binary
    elif output == 'raw_hex':
        return as_binary.hex()
    elif output in ['hex', 'prefix_hex']:
        return '0x' + as_binary.hex()
    else:
        raise Exception('unknown output format: ' + str(output))
